// *** Lucifer any platform ***
// Update of Windows CE Version 1.2 to make it platform-independant.
// Franois PESSAUX September 2007
// This program is freeware.
// Use its sources as you whant provided that you leave this copyright unchanged
// and that you drop a small word telling that I'm the author of these sources.
// This program is heavily based on
// David A. Barrett's implementation od Cipher 3.0.
// (barrett@boulder.Colorado.EDU). Thank's to him for having provided his code !
// Thank's a lot.
//   --Francois


#include <stdlib.h>	/* AES: getopt ANSII C: exit */
#include <stdio.h>
#include <string.h>
#include <libgen.h>     /* For basename. */
#include "cipher.h"

#define MAX_STRINGLEN 256

enum crypt_action { CA_encrypt, CA_decrypt } ;


#define	IO_BUFSIZE	1024		/* small for data cache speed */
#define KEYBUFLEN	129		/* allow long keys */


typedef struct {
  char	   iv[DES_BLOCKSIZE];
  keyType key;
  int	   ivlen;
  int	   decr;
} KeyItem;

typedef unsigned char block[DES_BLOCKSIZE];
/*
 * Keys which could accidently reproduce the plaintext with multiple-encryption
 */
block badKeys[16] = {
  { 0xe0, 0x00, 0xe0, 0x00, 0xf0, 0x00, 0xf0, 0x00 },
  { 0x00, 0xe0, 0x00, 0xe0, 0x00, 0xf0, 0x00, 0xf0 }, 
  { 0xfe, 0x1e, 0xfe, 0x1e, 0xfe, 0x0e, 0xfe, 0x0e },
  { 0x1e, 0xfe, 0x1e, 0xfe, 0x0e, 0xfe, 0x0e, 0xfe }, 
  { 0xe0, 0x1e, 0xe0, 0x1e, 0xf0, 0x0e, 0xf0, 0x0e },
  { 0x1e, 0xe0, 0x1e, 0xe0, 0x0e, 0xf0, 0x0e, 0xf0 }, 
  { 0x00, 0xfe, 0x00, 0xfe, 0x00, 0xfe, 0x00, 0xfe },
  { 0xfe, 0x00, 0xfe, 0x00, 0xfe, 0x00, 0xfe, 0x00 },
  { 0x00, 0x1e, 0x00, 0x1e, 0x00, 0x0e, 0x00, 0x0e },
  { 0x1e, 0x00, 0x1e, 0x00, 0x0e, 0x00, 0x0e, 0x00 }, 
  { 0xe0, 0xfe, 0xe0, 0xfe, 0xf0, 0xfe, 0xf0, 0xfe },
  { 0xfe, 0xe0, 0xfe, 0xe0, 0xfe, 0xf0, 0xfe, 0xf0 }, 
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },	/* self-dual */
  { 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe, 0xfe },	/* self-dual */
  { 0xe0, 0xe0, 0xe0, 0xe0, 0xf0, 0xf0, 0xf0, 0xf0 },	/* self-dual */
  { 0x1e, 0x1e, 0x1e, 0x1e, 0x0e, 0x0e, 0x0e, 0x0e },	/* self-dual */
};




/*
 * Return nonzero if the key is one of the well-known "weak" keys (FIPS Pub 74)
 */
int weakKey (char *keybits)
{
  block *keyp = &(badKeys[0]) ;	/* common compiler bug here */
  unsigned count = sizeof(badKeys) / sizeof(badKeys[0]) ;
  unsigned i ;

  do {
    for (i = 0; i < DES_BLOCKSIZE; i++) {
      if ((((*keyp)[i] ^ keybits[i]) & 0xfe) != 0) goto notfound ;
    }
    break ;
  notfound:
    keyp++ ;
  } while (--count != 0) ;

  return (count) ;
}



/*
 * Because we are using CFB, we always encrypt.  Since we'd like to be
 * able to run the chips in either mode, perhaps CBC would be better.
 * If CBC is used though, the last block must be handled in OFB mode.  Sigh.
 *
 * Returns:  
 *  -1 - out of memory (program exits, and error reported)
 *   0 - success
 *   1 - key is a weak key
 */
int setKey (keyType *kp, char *keybits, char *keystr, int mode)
{
  int res ;

  if (keystr != (char *) 0) {
    desKey (keybits, keystr, ' ') ;		/* space padded */
    memset (keystr, '\0', strlen (keystr)) ;	/* blank out key asap */
  }

  res = weakKey (keybits) ;

  makeKey (kp, keybits, DES_BLOCKSIZE, mode) ;	/* 0 means encrypt mode */
  memset (keybits, '\0', DES_BLOCKSIZE) ;  	/* zero out bits asap */

  if (*kp == (keyType) 0) {
    // Couldn't allocate memory for encryption key
    return (-1) ;
  }

  return (res) ;
}



/* Means that file over-writting if forbidden. Can be toggled */
/* to enable with the -f option.                              */
unsigned char force_flag = 0 ;



int EDCryption (enum crypt_action action, char *in_file_name, char *keybuf)
{
  KeyItem key_item ;
  char bufs[2][IO_BUFSIZE] ;
  char *chp, *ibuf = bufs[1] ;
  char *obuf = bufs[0] ;
  char keybits[DES_BLOCKSIZE] ;
  char out_file_name[MAX_STRINGLEN] ;
  int i ;
  unsigned long rcount, wcount ;   /* read/write counters. */
  FILE *in_file = NULL ;
  FILE *out_file = NULL ;

  in_file = fopen (in_file_name, "rb") ;
  if (in_file == NULL) {
    fprintf (stderr, "Unable to open input file.\n") ;
    return (-1) ;
  }

  /* Generate the file name according if we are */
  /* in an encryption or decryption mode.       */
  switch (action) {
  case CA_decrypt :
    /* Decryption. So extract the output file name from the archive. */
    /* First the original file name length EXCLUDING the final `\0`. */
    fread (&i, sizeof (int), 1, in_file) ;
    fread (out_file_name, sizeof (char), i, in_file) ;
    out_file_name[i] = '\0' ;
    break ;
  case CA_encrypt:
    /* Encryption. So file name is the current filename + ".luc" extension. */
    strcpy (out_file_name, in_file_name) ;
    strcat (out_file_name, ".luc") ;
    break ;
  }

  /* Test if the file already exists. */
  out_file = fopen (out_file_name, "rb") ;
  if ((out_file != NULL) && (!force_flag)) {
    /* The file already exists. */
    fclose (out_file) ;
    fprintf (stderr, "A file with this name already exists. Use -f to override.\n") ;
    fclose (in_file) ;
    return (-1) ;
  }

  /* Now finally create the output file. */
  out_file = fopen (out_file_name, "wb") ;
  if (out_file == NULL) {
    fprintf (stderr, "Unable to create out file.\n") ;
    fclose (in_file) ;
    return (-1) ;
  }

  /* And finally perform the great job... */
  if (action == CA_encrypt) {
    /* Attention : some implementations of basename may modify the passed */
    /* string. So we will us a copy of the original in_file_name.         */
    char tmp_filename [MAX_STRINGLEN] ;
    char *base_name ;
    strcpy (tmp_filename, in_file_name) ;

    /* Encryption. So first store the length of the original "basename" */
    /* (i.e filename without its path) EXCLUDING the final `\0`.        */
    base_name = basename (tmp_filename) ;
    i = strlen (base_name) ;
    fwrite (&i, sizeof (int), 1, out_file) ;
    fwrite (base_name, sizeof (char), i, out_file) ;
  }

  /* Generate the encryption/decryption key. */
  /* FALSE = encrypt, TRUE = decrypt. */
  key_item.decr = ((action == CA_encrypt) ? 0 : 1) ;
  key_item.ivlen = DES_BLOCKSIZE ;
  key_item.key = NULL ;                        /* Must be NULL !!! */
  memset (key_item.iv, '\0', DES_BLOCKSIZE) ;
  i = setKey (&key_item.key, keybits, keybuf, 0) ;

  // We disallow weak keys only during encryption in case where the
  // file was encrypted by another compatible program that allowed
  // encryption with a weak key.
  if (i && (action == CA_encrypt)) {
    fprintf (stderr, "Weak key used. Encryption rejected.\n") ;
    desDestroyKey (&key_item.key) ;
    fclose (in_file) ;
    fclose (out_file) ;
    return (-1) ;
  }

  do {
    
    rcount = fread (obuf, sizeof (char), IO_BUFSIZE, in_file) ;
    if (rcount <= 0) break ;
    chp = obuf ;
    obuf = ibuf ;
    ibuf = chp ;
    key_item.ivlen =
      desCFB (obuf, ibuf, rcount, key_item.iv, key_item.ivlen,
	      key_item.key, key_item.decr) ;
    wcount = fwrite (obuf, sizeof (char), rcount, out_file) ;
  } while (wcount == rcount) ;

  /* Perform the cleanup ASAP. */
  desDestroyKey (&key_item.key) ;
  memset (bufs[0], 0, IO_BUFSIZE) ;
  memset (bufs[1], 0, IO_BUFSIZE) ;
  fclose (in_file) ;
  fclose (out_file) ;
  /* Means success. */
  return (0) ;
}





int CheckAuthentification (enum crypt_action action, char *file_name)
{
  int ret_code  = -1 ;
  char passwd1[KEYBUFLEN] ;
  char passwd2[KEYBUFLEN] ;

  /* Ask the password a first time. */
  printf ("Enter password: ") ;
  fflush (stdout) ;
  scanf ("%s", passwd1) ;
  if (passwd1[0] == '\0') {
    fprintf (stderr, "Empty password. Encryption aborted.\n") ;
    goto cleanup ;
  }
	
  if (action == CA_encrypt) {
    /* Ask again the password for confirmation. */
    printf ("Enter again password (confirmation): ") ;
    fflush (stdout) ;
    scanf ("%s", passwd2) ;
    /* Now test that the 2 passwords match. */
    if (strcmp (passwd1, passwd2) != 0) {
      fprintf (stderr, "Passwords do not match. Encryption aborted.\n") ;
      goto cleanup ;
    }
  }
  
  /* Clear the second password because useless now. */
  memset (passwd2, 0, KEYBUFLEN * sizeof (char)) ;
  ret_code = EDCryption (action, file_name, passwd1) ;

 cleanup:
  // Clear the ASCII passwords
  memset (passwd1, 0, KEYBUFLEN * sizeof (char)) ;
  memset (passwd2, 0, KEYBUFLEN * sizeof (char)) ;

  return (ret_code) ;
}



void print_help ()
{
  printf ("Lucifer any platform.\nUpdate of Windows CE Version 1.2 to make it platform-independant.\nBy Franois PESSAUX September 2007.\nThis program is freeware.\nThis program is heavily based on David A. Barrett's implementation od Cipher 3.0 (barrett@boulder.Colorado.EDU). Thank's to him for having provided his code !\n\n") ;
  printf ("In encoding mode, Lucifer takes an input file and crypts it, embedding the original file name and storing in a file with the original filename + the \".luc\" extention.\n") ;
  printf ("In decoding mode, Lucifer takes an input file and decrypts it, storing the result in the originally embedded file name.\n\n") ;
  printf ("Usage: lucifer {-defh} file\n") ;
  printf ("Options:\n") ;
  printf ("\td : Decrypt.\n") ;
  printf ("\te : Encrypt.\n") ;
  printf ("\tf : Force file over-writing.\n") ;
  printf ("\th : Print this help.\n") ;
}



extern int optind ;
extern char *optarg ;


int main (int argc, char *argv[])
{
  int ch ;
  int num_actions_seen = 0 ;
  enum crypt_action action ;

  while ((ch = getopt(argc, argv, "defh")) != EOF) {
    switch (ch) {
    case 'd': action = CA_decrypt ; num_actions_seen++ ; break ;
    case 'e': action = CA_encrypt ; num_actions_seen++ ; break ;
    case 'f': force_flag = 1 ; break ;
    case 'h':
      print_help () ;
      exit (0) ;
      break ;
    default:
      fprintf (stderr, "Usage: {-defh}\n");
      exit (0) ;
      break ;
    }
  }
  argc -= optind ;
  argv += optind ;

  switch (num_actions_seen) {
  case 0:
    fprintf (stderr, "Missing action. Should be either -e or -d.\n" ) ;
    exit (-1) ;
    break ;

  case 1:
    if (argc != 1) {
      char tmp_input_name[MAX_STRINGLEN] ;
      /* fprintf (stderr, "Missing input file.\n") ; */
      fprintf (stdout, "Missing input file. Enter input file name: ") ;
      fflush (stdout) ;
      scanf ("%s", tmp_input_name) ;
      /* If an error occured, show an abnormal exit code. */
      if ((CheckAuthentification (action, tmp_input_name)) != 0) exit (-1) ;
    }
    else {
      /* If an error occured, show an abnormal exit code. */
      if ((CheckAuthentification (action, argv[0])) != 0) exit (-1) ;
    }
    break ;

  default:
    fprintf (stderr, "Multiple actions. Should be either -e or -d.\n" ) ;
    exit (-1) ;
    break ;
  }

  exit (0) ;
}
